# ScintillaTK class bindings
#
# Copyright (c) 2013 Mentor Graphics Corporation
# The license.terms file describes the conditions under which this software 
# may be distributed.
#

option add *Scintilla.textbackground white

# We won't assume clients want a default selection action when a multi-click
# event occurs in the margin area - we'll leave that up to the client
bind Scintilla <Double-1> {
    if {[%W margin index @%x,%y] == ""} {
        set tk::Priv(selectMode) word
        tk::TextSelectTo %W %x %y
        catch {%W mark set insert sel.first}
    }
}
bind Scintilla <Triple-1> {
    if {[%W margin index @%x,%y] == ""} {
        set tk::Priv(selectMode) line
        tk::TextSelectTo %W %x %y
        catch {%W mark set insert sel.first}
    }
}

bind Scintilla <Shift-1> {
    if {[%W margin index @%x,%y] == ""} {
        tk::TextResetAnchor %W @%x,%y
        set tk::Priv(selectMode) char
        tk::TextSelectTo %W %x %y
        catch {%W mark set insert sel.first}
    }
}

bind Scintilla <Double-Shift-1>  {
    if {[%W margin index @%x,%y] == ""} {
        set tk::Priv(selectMode) word
        tk::TextSelectTo %W %x %y 1
        catch {%W mark set insert sel.first}
    }
}

bind Scintilla <Triple-Shift-1>  {
    if {[%W margin index @%x,%y] == ""} {
        set tk::Priv(selectMode) line
        tk::TextSelectTo %W %x %y
        catch {%W mark set insert sel.first}
    }
}

bind Scintilla <Control-1> {
    if {[%W margin index @%x,%y] == ""} {
        catch {%W mark set insert @%x,%y}
    }
}

# Mimic some of the button bindings from Tk Text widget
set btns {
    1
    B1-Motion
    B1-Enter
    B1-Leave
    ButtonRelease-1
}
foreach e $btns {
    bind Scintilla <$e> [bind Text <$e>]
}

#OVERLOAD: we have to overload this one because of the -state value check where
# the Tk code looks for "== normal". Since our Scintilla widget implementation
# allows a state value of "readonly" and we still want to be able to receive
# focus upon a click, we change the check to "!= disabled".
proc ::tk::TextButton1 {w x y} {
    variable ::tk::Priv

    set Priv(selectMode) char
    set Priv(mouseMoved) 0
    set Priv(pressX) $x
    set anchorname [tk::TextAnchor $w]
    $w mark set insert [TextClosestGap $w $x $y]
    $w mark set $anchorname insert
    # Set the anchor mark's gravity depending on the click position
    # relative to the gap
    set bbox [$w bbox [$w index $anchorname]]
    if {$x > [lindex $bbox 0]} {
        $w mark gravity $anchorname right
    } else {
        $w mark gravity $anchorname left
    }
    # Allow focus in any case on Windows, because that will let the
    # selection be displayed even for state disabled text widgets.
    if {$::tcl_platform(platform) eq "windows" \
            || [$w cget -state] ne "disabled"} {
        focus $w
    }
}

# Like TextButton1, we change the logic checking the state
bind Scintilla <Tab> {
    if {[%W cget -state] ne "disabled"} {
        ::tk::ScintillaInsert %W \t
        focus %W
        break
    }
}

# Note: this is an overload of the function called by the Text class
# binding for B1-Leave. We're using the pre-8.5 implementation because
# beginning in 8.5 the scrolling was pixel-based, which isn't supported
# in Scintilla.
proc ::tk::TextAutoScan {w} {
    variable ::tk::Priv
    if {![winfo exists $w]} return
    if {$Priv(y) >= [winfo height $w]} {
        $w yview scroll 2 units
    } elseif {$Priv(y) < 0} {
        $w yview scroll -2 units
    } elseif {$Priv(x) >= [winfo width $w]} {
        $w xview scroll 2 units
    } elseif {$Priv(x) < 0} {
        $w xview scroll -2 units
    } else {
        return
    }
    TextSelectTo $w $Priv(x) $Priv(y)
    set Priv(afterId) [after 50 [list tk::TextAutoScan $w]]
}

#OVERLOAD: see below
proc ::tk::TextKeySelect {w new} {

    set anchorname [tk::TextAnchor $w]
    if {[$w tag nextrange sel 1.0 end] eq ""} {
        set insert_ix [$w index insert]
        if {[$w compare $new < insert]} {
            $w tag add sel $new insert
        } else {
            $w tag add sel insert $new
        }
        #----------------------------------------------------------------------
        # This differs from the code in text.tcl because of a fundamental
        # difference between Scintilla selection behavior and Tk Text.
        # The latter allows independent positioning of the insert location
        # relative to the selection; whereas Scintilla requires it to be
        # one of the extents of the selection region. So the problem we face,
        # which is the reason for this overload, is the above call to add
        # the selection: "tag add sel insert $new" will actually cause a change
        # in the location of the insertion point with Scintilla. Since the
        # intent of the next command is to move the anchor to where the insert
        # point was at the start of this function, we get that location first
        # and then reuse it via "$insert_ix"
        #----------------------------------------------------------------------
        $w mark set $anchorname $insert_ix
    } else {
        if {[$w compare $new < $anchorname]} {
            set first $new
            set last $anchorname
        } else {
            set first $anchorname
            set last $new
        }
        $w tag remove sel 1.0 $first
        $w tag add sel $first $last
        $w tag remove sel $last end
    }
    $w mark set insert $new
    $w see insert
    update idletasks
}




# define the class key bindings
bind Scintilla <Left>         { %W mark set insert "charleft" }
bind Scintilla <Right>        { %W mark set insert "charright" }
bind Scintilla <Up>           { %W mark set insert "lineup" }
bind Scintilla <Down>         { %W mark set insert "linedown" }
bind Scintilla <Home>         { %W mark set insert "linehome" }
bind Scintilla <End>          { %W mark set insert "lineend" }
bind Scintilla <Control-Home> { %W mark set insert "top" }
bind Scintilla <Control-End>  { %W mark set insert "bottom" }

# Mimic the key bindings from Tk Text widget
set keys {
    Control-a
    Control-b
    Control-backslash
    Control-d
    Control-Down
    Control-e
    Control-f
    Control-h
    Control-i
    Control-k
    Control-Left
    Control-n
    Control-Next
    Control-o
    Control-p
    Control-Prior
    Control-Right
    Control-Shift-End
    Control-Shift-Home
    Control-Shift-space
    Control-Shift-Tab
    Control-slash
    Control-t
    Control-v
    Control-Up
    Control-Tab
    Control-space
    Meta-b
    Meta-d
    Meta-f
    Meta-less
    Meta-greater
    Meta-backspace
    Meta-delete
    Next
    Prior
    Select
    Shift-Prior
    Shift-Next
    Shift-Control-Left
    Shift-Control-Right
    Shift-Control-Up
    Shift-Control-Down
    Shift-Down
    Shift-End
    Shift-Home
    Shift-Left
    Shift-Right
    Shift-Tab
    Shift-Up
    Shift-Select
}    
foreach e $keys {
    bind Scintilla <$e> [bind Text <$e>]
}

bind Scintilla <Return> {
   ::tk::ScintillaInsert %W \n
}

bind Scintilla <Delete> {
    if {[%W tag nextrange sel 1.0 end] ne ""} {
       %W delete sel.first sel.last
    } else {
       %W delete insert
       %W see insert
    }
}

bind Scintilla <BackSpace> {
    if {[%W tag nextrange sel 1.0 end] ne ""} {
       %W delete sel.first sel.last
    } elseif {[%W compare insert != 1.0]} {
       %W delete insert-1c
       %W see insert
    }
}

bind Scintilla <Control-KP_Add>      {%W zoomin}
bind Scintilla <Control-KP_Subtract> {%W zoomout}
if { [tk windowingsystem] eq "win32" } {
bind Scintilla <Control-Key-plus>    {%W zoomin}
bind Scintilla <Control-Key-minus>   {%W zoomout}
}

bind Scintilla <Insert> {
    catch {::tk::ScintillaInsert %W [::tk::GetSelection %W PRIMARY]}
}

bind Scintilla <KeyPress> {
   ::tk::ScintillaInsert %W %A
}

# This next group of bindings mimics what is used for the Tk Text widget.
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Ignore all Alt, Meta, and Control keypresses unless explicitly bound.
# Otherwise, if a widget binding for one of these is defined, the
# <KeyPress> class binding will also fire and insert the character,
# which is wrong.  Ditto for <Escape>.
bind Scintilla <Alt-KeyPress>     {# nothing }
bind Scintilla <Meta-KeyPress>    {# nothing}
bind Scintilla <Control-KeyPress> {# nothing}
bind Scintilla <Escape>           {# nothing}
bind Scintilla <KP_Enter>         {# nothing}

bind Scintilla <<Redo>> {
    catch { %W edit redo }
}

bind Scintilla <<Undo>> {
    catch { %W edit undo }
}

# MOUSE WHEEL
bind Scintilla <Button-4> {
   %W yview scroll -5 units
}
bind Scintilla <Button-5> {
   %W yview scroll 5 units
}
if { [tk windowingsystem] eq "win32" } {
   # NOTE: match Tk 8.4 behavior, since 8.5 uses pixel-level scrolling
   bind Scintilla <MouseWheel> {
      %W yview scroll [expr {- (%D / 120) * 4}] units
   }
}

#############################################################################
# define the functions tied to the bindings
proc ::tk::ScintillaInsert {w s} {
   # prevent insert if state is disabled or readonly
   if {$s eq "" || [$w cget -state] ne "normal"} {
      return
   }
#puts "INSERT $w --> (($s))"
    if {[$w tag nextrange sel 1.0 end] ne ""} {
       $w delete sel.first sel.last
    }
   $w insert insert $s
   $w see insert
}
